"""Zero and few-shot classification using SetFit."""
from typing import Iterable

import datasets
import numpy as np
from setfit import SetFitModel, SetFitTrainer, get_templated_dataset
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import NotFittedError

__all__ = ["SetFitZeroShotClassifier", "SetFitFewShotClassifier"]


class SetFitZeroShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible zero shot classification
    with SetFit and sentence transformers.

    Parameters
    ----------
    model_name: str, default 'sentence-transformers/all-MiniLM-L6-v2'
        Name of sentence transformer on HuggingFace Hub.
    sample_size: int, default 8
        Number of training samples to generate.
    device: str, default 'cpu'
        Indicates which device should be used for classification.
        Models are by default run on CPU.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
        sample_size: int = 8,
        device: str = "cpu",
    ):
        self.model_name = model_name
        self.classes_ = None
        self.model = SetFitModel.from_pretrained(model_name)
        self.model.model_body.to(device)
        self.trainer = None
        self.sample_size = sample_size

    def _train(self):
        if self.classes_ is None:
            raise NotFittedError(
                "You should collect all class labels before calling _train()"
            )
        train_dataset = get_templated_dataset(
            candidate_labels=list(self.classes_), sample_size=self.sample_size
        )
        self.trainer = SetFitTrainer(
            model=self.model, train_dataset=train_dataset
        )
        self.trainer.train()

    def fit(self, X, y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.classes_ = np.array(list(set(y)))
        self.n_classes = len(self.classes_)
        self._train()
        return self

    def partial_fit(self, X, y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        if self.classes_ is None:
            self.classes_ = np.array(list(set(y)))
            self._train()
        else:
            new_labels = set(self.classes_) - set(y)
            if new_labels:
                self.classes_ = np.concatenate(self.classes_, list(new_labels))
                self._train()
        self.n_classes = len(self.classes_)
        return self

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        X = list(X)
        if self.classes_ is None:
            raise NotFittedError(
                "Class labels have not been collected yet, please fit."
            )
        pred_ind = self.model(X)
        pred = self.classes_[pred_ind]
        # Produces a single str when one example is passed
        # and we do not like that.
        if isinstance(pred, str):
            return np.array([pred])
        return pred


class SetFitFewShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible few shot classification
    with SetFit and sentence transformers.

    Parameters
    ----------
    model_name: str, default 'sentence-transformers/all-MiniLM-L6-v2'
        Name of sentence transformer on HuggingFace Hub.
    device: str, default 'cpu'
        Indicates which device should be used for classification.
        Models are by default run on CPU.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
        device: str = "cpu",
    ):
        self.model_name = model_name
        self.classes_ = None
        self.model = SetFitModel.from_pretrained(model_name)
        self.model.model_body.to(device)
        self.trainer = None

    def fit(self, X: Iterable[str], y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: iterable of str
            Examples to pass into the few-shot prompt.
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        X = list(X)
        y = list(y)
        self.classes_ = np.array(list(set(y)))
        self.n_classes = len(self.classes_)
        train_dataset = datasets.Dataset.from_dict(dict(text=X, label=y))
        self.trainer = SetFitTrainer(
            model=self.model, train_dataset=train_dataset
        )
        self.trainer.train()
        return self

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        X = list(X)
        if self.trainer is None:
            raise NotFittedError("You need to fit the model before inference.")
        return self.model(X)



"""Zero shot classification with text-to-text language models."""
from typing import Iterable

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import NotFittedError
from thefuzz import process
from tqdm import tqdm
from transformers import AutoModelForSeq2SeqLM, AutoTokenizer

__all__ = ["Text2TextZeroShotClassifier", "Text2TextFewShotClassifier"]

default_prompt = """
I will give you a piece of text. Please classify it
as one of these classes: {classes}. Please only respond
with the class label in the same format as provided.

'{X}'
"""


class Text2TextZeroShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible zero shot classification
    with seq2seq language models.

    Parameters
    ----------
    model_name: str, default 'google/flan-t5-base'
        Text-to-text instruct model from HuggingFace.
    prompt: str, optional
        You can specify the prompt which will be used to prompt the model.
        Use placeholders to indicate where the class labels and the
        data should be placed in the prompt.
    max_new_tokens: int, default 256
        Maximum number of tokens the model should generate.
    fuzzy_match: bool, default True
        Indicates whether the output lables should be fuzzy matched
        to the learnt class labels.
        This is useful when the model isn't giving specific enough answers.
    progress_bar: bool, default True
        Indicates whether a progress bar should be shown.
    device: str, default 'cpu'
        Indicates which device should be used for classification.
        Models are by default run on CPU.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        model_name: str = "google/flan-t5-base",
        prompt: str = default_prompt,
        max_new_tokens: int = 256,
        fuzzy_match: bool = True,
        progress_bar: bool = True,
        device: str = "cpu",
    ):
        self.model_name = model_name
        self.prompt = prompt
        self.device = device
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name).to(
            self.device
        )
        self.classes_ = None
        self.progress_bar = progress_bar
        self.max_new_tokens = max_new_tokens
        self.fuzzy_match = fuzzy_match

    def fit(self, X, y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.classes_ = np.array(list(set(y)))
        self.n_classes = len(self.classes_)
        return self

    def partial_fit(self, X, y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        if self.classes_ is None:
            self.classes_ = np.array(list(set(y)))
        else:
            new_labels = set(self.classes_) - set(y)
            if new_labels:
                self.classes_ = np.concatenate(self.classes_, list(new_labels))
        self.n_classes = len(self.classes_)
        return self

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        if self.classes_ is None:
            raise NotFittedError(
                "Class labels have not been collected yet, please fit."
            )
        pred = []
        if self.progress_bar:
            X = tqdm(X)
        for text in X:
            classes_in_quotes = [f"'{label}'" for label in self.classes_]
            inputs = self.tokenizer(
                self.prompt.format(
                    X=text, classes=", ".join(classes_in_quotes)
                ),
                return_tensors="pt",
            ).to(self.device)
            output = self.model.generate(
                **inputs, max_new_tokens=self.max_new_tokens
            )
            label = self.tokenizer.decode(output[0], skip_special_tokens=True)
            if self.fuzzy_match and label not in self.classes_:
                label, _ = process.extractOne(label, self.classes_)
            pred.append(label)
        return np.array(pred)


fewshot_prompt = """
I will give you a piece of text. Please classify it
as one of these classes: {classes}. Please only respond
with the class label in the same format as provided.
Here are some examples of texts labelled by experts.
{examples}
Label this piece of text:
'{X}'
"""

example_prompt = """
Examples of texts labelled '{label}':
{examples}
"""


class Text2TextFewShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible few-shot classification
    with seq2seq language models.

    Parameters
    ----------
    model_name: str, default 'google/flan-t5-base'
        Text-to-text instruct model from HuggingFace.
    prompt: str, optional
        You can specify the prompt which will be used to prompt the model.
        Use placeholders to indicate where the class labels and the
        data should be placed in the prompt.
    max_new_tokens: int, default 256
        Maximum number of tokens the model should generate.
    fuzzy_match: bool, default True
        Indicates whether the output lables should be fuzzy matched
        to the learnt class labels.
        This is useful when the model isn't giving specific enough answers.
    progress_bar: bool, default True
        Indicates whether a progress bar should be shown.
    device: str, default 'cpu'
        Indicates which device should be used for classification.
        Models are by default run on CPU.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    examples_: dict of str to list of str
        Learned examples for each class.
    """

    def __init__(
        self,
        model_name: str = "google/flan-t5-base",
        prompt: str = default_prompt,
        max_new_tokens: int = 256,
        fuzzy_match: bool = True,
        progress_bar: bool = True,
        device: str = "cpu",
    ):
        self.model_name = model_name
        self.prompt = prompt
        self.device = device
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name).to(
            self.device
        )
        self.classes_ = None
        self.progress_bar = progress_bar
        self.max_new_tokens = max_new_tokens
        self.fuzzy_match = fuzzy_match
        self.examples_: dict[str, list[str]] = dict()

    def fit(self, X: Iterable[str], y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: iterable of str
            Examples to pass into the few-shot prompt.
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.examples_ = dict()
        for text, label in zip(X, y):
            if label not in self.examples_:
                self.examples_[label] = []
            self.examples_[label].append(text)
        self.classes_ = np.array(list(self.examples_.keys()))
        self.n_classes = len(self.classes_)
        return self

    def partial_fit(self, X: Iterable[str], y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: iterable of str
            Examples to pass into the few-shot prompt.
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        for text, label in zip(X, y):
            if label not in self.examples_:
                self.examples_[label] = []
            self.examples_[label].append(text)
        self.classes_ = np.array(list(self.examples_.keys()))
        self.n_classes = len(self.classes_)
        return self

    def generate_prompt(self, text: str) -> str:
        """Generates prompt based on the model's parameters."""
        if self.examples_ is None:
            raise NotFittedError(
                "No examples have been learnt yet, fit the model."
            )
        text_examples = []
        for label, examples in self.examples_.items():
            examples = [f"'{example}'" for example in examples]
            subprompt = example_prompt.format(
                label=f"'{label}'", examples="\n".join(examples)
            )
            text_examples.append(subprompt)
        examples_subprompt = "\n".join(text_examples)
        classes_in_quotes = [f"'{label}'" for label in self.classes_]
        prompt = fewshot_prompt.format(
            X=text,
            classes=", ".join(classes_in_quotes),
            examples=examples_subprompt,
        )
        return prompt

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        if self.classes_ is None:
            raise NotFittedError(
                "Class labels have not been collected yet, please fit."
            )
        pred = []
        if self.progress_bar:
            X = tqdm(X)
        for text in X:
            prompt = self.generate_prompt(text)
            inputs = self.tokenizer(
                prompt,
                return_tensors="pt",
            ).to(self.device)
            output = self.model.generate(
                **inputs, max_new_tokens=self.max_new_tokens
            )
            label = self.tokenizer.decode(output[0], skip_special_tokens=True)
            if self.fuzzy_match and label not in self.classes_:
                label, _ = process.extractOne(label, self.classes_)
            pred.append(label)
        return np.array(pred)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stormtrooper']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23.0,<2.0.0',
 'scikit-learn>=1.2.0,<2.0.0',
 'thefuzz>=0.18.0,<0.19.0',
 'tqdm>=4.60.0,<5.0.0',
 'transformers>=4.25.0,<5.0.0']

extras_require = \
{'setfit': ['setfit>=0.7.0,<0.8.0', 'datasets>=2.14.0,<3.0.0'],
 'torch': ['torch>=2.0.0,<3.0.0']}

setup_kwargs = {
    'name': 'stormtrooper',
    'version': '0.3.0',
    'description': 'Transformer-based zero and few-shot classification in scikit-learn pipelines',
    'long_description': '<img align="left" width="82" height="82" src="assets/logo.svg">\n\n# stormtrooper\n\n<br>\nTransformer-based zero/few shot learning components for scikit-learn pipelines.\n\n[Documentation](https://centre-for-humanities-computing.github.io/stormtrooper/)\n\n## Example\n\n```bash\npip install stormtrooper\n```\n\n```python\nclass_labels = ["atheism/christianity", "astronomy/space"]\nexample_texts = [\n    "God came down to earth to save us.",\n    "A new nebula was recently discovered in the proximity of the Oort cloud."\n]\n```\n\n\n### Zero-shot learning\n\nFor zero-shot learning you can use zero-shot models:\n```python\nfrom stormtrooper import ZeroShotClassifier\nclassifier = ZeroShotClassifier().fit(None, class_labels)\n```\n\nGenerative models (GPT, Llama):\n```python\nfrom stormtrooper import GenerativeZeroShotClassifier\n# You can hand-craft prompts if it suits you better, but\n# a default prompt is already available\nprompt = """\n### System:\nYou are a literary expert tasked with labeling texts according to\ntheir content.\nPlease follow the user\'s instructions as precisely as you can.\n### User:\nYour task will be to classify a text document into one\nof the following classes: {classes}.\nPlease respond with a single label that you think fits\nthe document best.\nClassify the following piece of text:\n\'{X}\'\n### Assistant:\n"""\nclassifier = GenerativeZeroShotClassifier(prompt=prompt).fit(None, class_labels)\n```\n\nText2Text models (T5):\nIf you are running low on resources I would personally recommend T5.\n```python\nfrom stormtrooper import Text2TextZeroShotClassifier\n# You can define a custom prompt, but a default one is available\nprompt = "..."\nclassifier =Text2TextZeroShotClassifier(prompt=prompt).fit(None, class_labels)\n```\n\n```python\npredictions = classifier.predict(example_texts)\n\nassert list(predictions) == ["atheism/christianity", "astronomy/space"]\n```\n\n### Few-Shot Learning\n\nFor few-shot tasks you can only use Generative and Text2Text (aka. promptable) models.\n\n```python\nfrom stormtrooper import GenerativeFewShotClassifier, Text2TextFewShotClassifier\n\nclassifier = Text2TextFewShotClassifier().fit(example_texts, class_labels)\npredictions = model.predict(["Calvinists believe in predestination."])\n\nassert list(predictions) == ["atheism/christianity"]\n```\n\n### Fuzzy Matching\n\nModels by default will fuzzy match results to the closest class label, you can disable this behavior\nby specifying `fuzzy_match=False`.\n\nIf you want fuzzy matching speedup, you should install `python-Levenshtein`.\n\n### Inference on GPU\n\nFrom version 0.2.2 you can run models on GPU.\nYou can specify the device when initializing a model:\n\n```python\nclassifier = Text2TextZeroShotClassifier(device="cuda:0")\n```\n',
    'author': 'Márton Kardos',
    'author_email': 'power.up1163@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from stormtrooper.generative import (
    GenerativeFewShotClassifier,
    GenerativeZeroShotClassifier,
)
from stormtrooper.text2text import (
    Text2TextFewShotClassifier,
    Text2TextZeroShotClassifier,
)
from stormtrooper.zero_shot import ZeroShotClassifier

__all__ = [
    "GenerativeZeroShotClassifier",
    "GenerativeFewShotClassifier",
    "Text2TextZeroShotClassifier",
    "Text2TextFewShotClassifier",
    "ZeroShotClassifier",
]
